//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                   (c) 1997 - 2010, European Space Agency
//
//---------------------------------------------------------------------------
//
//	ExprConstructor.cc:	--- implementation of ExpressionConstructor class
//
//	CHANGE HISTORY:
//		UL, 1998-11-05: 1.0.0: - first version
//		UL, 1998-04-27: 1.1.0: - added QTooltip to "clear" button
//		UL, 2000-02-25: 1.2.0: - appendExpression(): slightly altered logic
//								 to append sub-expressions; this fixes
//								 SPR 1205 : Incorrect selection expression
//		UL, 2000-05-10: 1.1.0: - constructor takes initial selection
//								 expression as argument (optional)
//		UL, 2000-07-05: 1.2.0: - replaced QLineEdit with QMultiLineEdit;
//								 the latter has on-demand scrollbars which
//								 closes SPR 1657: scroll bar wanted on
//								 selection expression
//		UL, 2000-11-07: 1.3.0: - various updates
//		UL, 2001-08-14: 1.4.0: - font size of expression string is no longer
//								 fixed to 12 pt but family is always "Fixed"
//								 (closes SPR 2123 "fonts")
//
//	AUTHOR:
//		U. Lammers (UL), Thu Nov  5 10:01:37 MET 1998
//
//	$Header: /usr/local/cvs/sasdev/task_devel/packages/xmmselect/src/ExprConstructor.cc,v 1.11 2002/07/11 16:08:13 ulammers Exp $
//
#include <fstream>
#include <cstring>
#include <qlabel.h>
#include <qmultilineedit.h>
#include <qmessagebox.h>
#include <qscrollbar.h>
#include <qpushbutton.h>
#include <qtooltip.h>

#include "XmmSelectServer.h"
#include "ExprConstructor.h"
#include "PreferenceMenu.h"
#include "DataSetBrowser.h"
#include "DataSubspace.h"

using namespace std;

ExpressionConstructor::ExpressionConstructor(QWidget *parent,
											 const char *name,
											 const string &expr)
	: QGroupBox(parent, name)
{
	this->setTitle("Selection expression");

	QBoxLayout *vbox = new QVBoxLayout(this, 20);
	QBoxLayout *hboxedit = new QHBoxLayout();
	vbox->addLayout(hboxedit, 0);

	_exprEdit = new QMultiLineEdit(this);
	_exprEdit->setMaxLines(3);
	fontChange(QFont());
	hboxedit->addWidget(_exprEdit, 1);
	if (expr.length())
		_exprEdit->setText(expr.c_str());
	QToolTip::add(_exprEdit, "The selection expression");

	QBoxLayout *hboxbutt = new QHBoxLayout();
	vbox->addLayout(hboxbutt, 0);

	//
	//	check if we have a PrefSelections in the preferences
	//
	const PreferenceGroup *pg = xmmSelectServer->preferences();
	if (pg->hasPreference(prefSelectionName)) {
		PreferenceButtonMenu *prefButt =
			new PreferenceButtonMenu(this,
					*pg->preference(prefSelectionName)->group());
		connect(prefButt->popup(),
				SIGNAL(preferenceValue(const std::string &,
									   const PreferenceValue &)),
				this,
				SLOT(prefSelSelected(const std::string &,
					 const PreferenceValue &)));

		QToolTip::add(prefButt, "Select expressions from preferences");
		hboxbutt->addWidget(prefButt);
//		hboxbutt->addStretch(1);
	}

	//
	//	add import button
	//
	QPushButton *importButt = new QPushButton("import", this);
	QToolTip::add(importButt, "Import selection expression from attribute "
				  "or DSS in block");
	connect(importButt, SIGNAL(released()), this, SLOT(import()));
	hboxbutt->addWidget(importButt);

	hboxbutt->addStretch(10);
	QPushButton *clearButt = new QPushButton("clear", this);
	QToolTip::add(clearButt, "Clear the selection expression");
	connect(clearButt, SIGNAL(released()), this, SLOT(clear()));
	clearButt->setMinimumSize(clearButt->sizeHint());
	hboxbutt->addWidget(clearButt);

	_exprCount = 0;
} // ExpressionConstructor::ExpressionConstructor()


void ExpressionConstructor::fontChange(const QFont &)
{
	QFont font = this->font();
	font.setFamily("Fixed");
	_exprEdit->setFont(font);
	_exprEdit->setMinimumHeight(3*_exprEdit->fontMetrics().height()+4);
	update();
} // ExpressionConstructor::fontChange()


string ExpressionConstructor::expression() const
{
	QString expr = _exprEdit->text().simplifyWhiteSpace();
	if (expr.isNull())
		return "";

	return string(expr);
} // ExpressionConstructor::expression()


void ExpressionConstructor::appendExpression(const string &expr,
											 AppendMode appendMode,
											 const string &sender)
{
	string expression(expression());
	if (expression != _lastExpression)
		_lastSender = "";

	if (expression.length() == 0) {
		_exprCount = 0;
		_lastSender = "";
		expression = "(" + expr;
	} else if (appendMode == OR ||
			   (appendMode == OR_IF_SAME_SENDER && sender == _lastSender))
		expression = expression.substr(0, expression.length()-1) +
					  "||" + expr;
	else
		expression += "&&(" + expr;

	expression += ")";

	_exprEdit->setText(expression.c_str());
	_lastSender = sender;
	_lastExpression = expression;
	++_exprCount;

} // ExpressionConstructor::appendExpression()


void ExpressionConstructor::manageSelectionExpressionDialog(const string &expr,
													   const string &message,
													   const string &title)
{

	SelectionExpressionDialog *diag = new SelectionExpressionDialog(
										xmmSelectServer->gui(),
										title.c_str(),
										message.c_str(),
										expr);
	assert(diag);
} // ExpressionConstructor::manageSelectionExpressionDialog()


void ExpressionConstructor::clear()
{
	_lastSender = "";
	_exprCount = 0;
	_exprEdit->setText(_lastSender.c_str());
} // ExpressionConstructor::clear()


void ExpressionConstructor::import()
{
	DataSetBrowser browser(xmmSelectServer->gui(),
						   "Select selection expression");
	browser.setMultiSelection(false);
//	browser.setSelectable(DalListViewable::Type(DalListViewable::BlockType|
//											  DalListViewable::AttributeType));
	browser.setSelectable(DalListViewable::Any);
	if (browser.exec() == DataSetBrowser::Accepted) {
		string fname(browser.selection()[0]);
		string expr;
		//	check if it is a set than is understood by the DAL
		if (dataSetServer->exists(fname)) {
			Block *block = dataSetServer->
					open(fname, Dal::Read, DataSetServer::Low)->block(fname);
			if (fname.rfind('%') != string::npos)
				//	select attribute
				expr = block->attribute(fname)->asString();
			else {
				//	a block - check if it has DSS
				DataSubspace dss(block);
				if (dss.components())
					expr = dss.selectExpression();
				else
					QMessageBox::warning(this, "No Data Subspace Information",
						"The specified block does not contain data subspace\n"
						"information.");
			}
			dataSetServer->close(block->parent());
			if (!expr.empty())
				prefSelSelected("Imported DSS expression",
								PreferenceValue(expr));

		} else {
			//	assume this is an ASCII file - read contents into expr string
			//	not more than 32k
			ifstream is(fname.c_str());
			char p[32768+1];
			std::memset(p, 0, sizeof(p));
			is.read(p, sizeof(p)-1);
			expr = p;
			std::string::size_type i = expr.find('\n');
			while (i != string::npos) {
				expr.replace(i, 1, " ");
				i = expr.find('\n');
			}
			if (!expr.empty())
				prefSelSelected("Imported selection expression",
								PreferenceValue(expr));
		}
	}
} // ExpressionConstructor::import()


void ExpressionConstructor::prefSelSelected(const std::string &text,
											const PreferenceValue &value)
{
	manageSelectionExpressionDialog(value, "", text);
} // ExpressionConstructor::prefSelSelected()


ExpressionConstructor::~ExpressionConstructor()
{
} // ExpressionConstructor::~ExpressionConstructor()


SelectionExpressionDialog::SelectionExpressionDialog(QWidget *parent,
													 const char *name,
													 const string &message,
													 const string &expr)
	: QDialog(parent, name, false)
{
	_vbox = new QVBoxLayout(this, 20);

	if (!message.empty()) {
		QLabel *label = new QLabel(message.c_str(), this);
		label->setFrameStyle(QFrame::Box | QFrame::Raised);
		_vbox->addWidget(label);
	}

	_hboxedit = new QHBoxLayout();
	_vbox->addLayout(_hboxedit, 5);

	_exprEdit = new QMultiLineEdit(this);
	QFont font = _exprEdit->font();
	font.setFamily("Fixed");
	_exprEdit->setFont(font);
	_exprEdit->setText(expr.c_str());
	_exprEdit->setMinimumWidth(500);
	_exprEdit->setMinimumHeight(_exprEdit->minimumSizeHint().height());
	_hboxedit->addWidget(_exprEdit);

	_hboxbutt = new	QHBoxLayout();
	_vbox->addLayout(_hboxbutt, 5);
	_hboxbutt->addStretch(1);
	_ok = new QPushButton("Import", this);
	connect(_ok, SIGNAL(released()), this, SLOT(accept()));
	_hboxbutt->addWidget(_ok);
	_hboxbutt->addStretch(1);
	_cancel = new QPushButton("Cancel", this);
	connect(_cancel, SIGNAL(released()), this, SLOT(reject()));
	_hboxbutt->addWidget(_cancel);
	_hboxbutt->addStretch(1);

	setCaption(name);
	show();

} // SelectionExpressionDialog::SelectionExpressionDialog()


void SelectionExpressionDialog::accept()
{
	xmmSelectServer->gui()->expressionConstructor()->appendExpression(
							string(_exprEdit->text()));
	delete this;
} // SelectionExpressionDialog::accept()


SelectionExpressionDialog::~SelectionExpressionDialog()
{
} // SelectionExpressionDialog::~SelectionExpressionDialog()
