//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	TaskLauncher.cc:	--- implementation of TaskLauncher classes
//
//	CHANGE HISTORY:
//		UL, 1998-10-22: 1.0.0: - first version
//		UL, 1999-03-01: 1.1.0: - many upgrades
//		UL, 1999-04-24: 1.1.1: - bug fix: before bringing up a viewer,
//								 inspect again the name of the output data set
//								 which might have been altered by the user
//								 in the parameter dialog
//		UL, 1999-08-31: 1.1.2: - dstoplot is now called as a function
//							   - create time series plots with TIME column
//		UL, 1999-09-20: 1.2.0: - consider binning factors in image extraction
//		UL, 1999-09-21: 1.3.0: - in addition to binning factors also retrieve
//								 image offset from evselect's parameters
//		UL, 1999-10-14: 1.4.0: - intercept invalid selection expression before
//								 evselect is run (uses new
//								 Evalulator::isValidExpression())
//							   - check errors from evselect (via exceptions)
//							   - pass file names with absolute paths to
//								 viewers (fixes SPR 854: ds9 reuses local file)
//		UL, 1999-10-22: 1.5.0: - set evselect's <timecolumn> parameter
//								 in rates curve extraction
//							   - bug fix in dsplot interface: "withx"/"withy"
//								 parameters were not set
//								 (fixes SPR 903: Light curve produced by
//								  XMMSELECT are displayed with inverted axis)
//		UL, 1999-20-28: 1.5.1: - do not read WCS keywords from X/Y images
//		UL, 2000-02-08: 1.6.0: - upon each invocation of evselect set
//								 parameters controlling generation of products
//								 back to default values
//								 (fixes SPR 1084: Easy to overwrite products
//								 without realising)
//		UL, 2000-05-09: 1.7.0: - minor fix
//		UL, 2000-06-05: 1.7.1: - d/f value of evselect's updateexposure is
//								 now 'false'
//		UL, 2000-06-20: 1.7.2: - invoke tasks through task() method instead
//								 of calling it directly
//		UL, 2000-07-26: 1.7.3: - minor bug fixes
//		UL, 2000-10-23: 1.8.0: - fixed SPR 1763: crash when making lightcurve
//								 with maketimecolumn unchecked
//							   - EvselectLauncher::_minX/_minY was incorrectly
//								 assigned in case of X/Y images; this fixes
//								 SPR 1743: xmmselect 2D regions from ds9 off
//								 by 0.5 pixel
//							   - invoke evselect with squarepixels=true by
//								 default
//		UL, 2000-10-31: 1.8.1: - better error message in dialog in case of
//								 evselect runtime error
//		UL, 2000-12-05: 1.8.2: - clear GUI status bar after evselect error
//							   - various changes to fix SPR 1875: Incorrect
//								 interval from grace
//		UL, 2000-12-06: 1.8.3: - close table after opening in
//								 isValidExpression(); this closes SPR 1879
//								 Cannot use current event list as output
//		UL, 2001-04-03: 1.9.0: - removed 'calculatebackscale' from
//								 evselect task invocation (following removal
//								 of this parameter in evselect-3.35)
//							   - fixed SPR 1992: incorrect spatial extraction
//								 for PMS data (_withMinX/Y was incorrectly
//								 set in case of X/Y images)
//		UL, 2001-05-15: 1.9.1: - invoke backscale with 'badpixplocation'
//								 set to set name (instead of event table name)
//		UL, 2001-05-17: 1.9.2: - implemented SPR 2046: Does not recover from
//								 error in backscale
//								 [try-catch block around backscale call]
//		UL, 2001-06-11: 1.9.3: - display extracted spectrum even if
//								 backscale has failed
//		UL, 2001-08-14: 1.9.4: - bug fix in conversion of ds9 region to
//								 column coordinates
//		UL, 2002-01-23: 1.10.0:- added support for evselect's 'makeratecolumn'
//								 parameter
//		UL, 2002-01-24: 1.11.0:- made 'makeratecolumn=true' the default;
//								 rates curve now plotted with error bars
//		UL, 2002-04-11: 1.12.0:- create "CHANNEL"-vs.-"RATE.STAT_ERR" plot
//								 when spectrum of rates was generated
//								 by evselect ("withzcolumn=true")
//		UL, 2002-05-17: 1.13.0:- impelement ImageDisplayLauncher::down()
//		UL, 2002-09-18: 1.14.0:- added implementation for new task wrappers
//								 (e.g. SourceRegionOptimizer)
//
//	AUTHOR:
//		U. Lammers (UL), Thu Oct 22 14:24:39 MET DST 1998
//
//	$Header: /usr/local/cvs/sasdev/task_devel/packages/xmmselect/src/TaskLauncher.cc,v 1.48 2002/09/18 15:24:16 ulammers Exp $
//
#include <string>
#include <sstream>
#include <cctype>
#include <unistd.h>

#include "ParamDialog.h"
#include "DataFilter.h"
#include "XmmSelectServer.h"
#include "TaskLauncher.h"
#include "common.h"
#include "param.h"

#include "metatask.h"

#include "Grace.h"
#include "dstoplot.h"
#include "evselect.h"
#include "imgdisplay.h"
#include "backscale.h"
#include "especget.h"
#include "eregionanalyse.h"
#include "Evaluator.h"

#include <qmessagebox.h>
#include <qpopupmenu.h>
#include <qdir.h>

using namespace std;

Grace* DsplotLauncher::_grace = 0;
double DsplotLauncher::_offsetX = 0.;

string DsplotLauncher::toAbsolutePath(const string &fname)
{
	string name(fname);
	if (name.length() && name[0] != '/') {
		//	relative path
		QDir cwd;
		name = string(cwd.absPath()) + "/" + fname;
	}

	return name;		
} // toAbsolutePath()


//	constructor
TaskLauncher::TaskLauncher(const string &name) : _name(name)
{
} // TaskLauncher::TaskLauncher()


TaskLauncher::~TaskLauncher()
{
} // TaskLauncher::~TaskLauncher()


//
//	----------------------( evselect )----------------------------------------
//
EvselectLauncher::EvselectLauncher()
	: TaskLauncher("evselect"), _quicklook(false), _withBackscaleDiag(false),
		_binX(1.), _binY(1.), _oneDbin(1.)
{
	const PreferenceGroup *pg = xmmSelectServer->preferences();
	if (pg->hasPreference("quicklook") &&
		pg->preference("quicklook")->hasValue())
		_quicklook = pg->preference("quicklook")->value().asBool();
	if (pg->hasPreference("withbackscalediag") &&
		pg->preference("withbackscalediag")->hasValue())
			_withBackscaleDiag = pg->preference("withbackscalediag")->value()
				.asBool();

	openParameters(name());

	//	set all product generation to false
	const char *boolParam[] = { "withfilteredset",
							    "keepfilteroutput",
							    "writedss",
								"updateexposure",
							    "withspectrumset",
							    "withimageset",
							    "withrateset",
							    "withhistogramset" };

	for (unsigned n=0; n<NUM_OF(boolParam); ++n)
		setParameter(boolParam[n], false);
} // EvselectLauncher::EvselectLauncher()


bool EvselectLauncher::filterTable(const string &dsName,
								   const string &expression,
								   const string &filteredSetName)
{
	if (!isValidExpression(expression))
		return false;

	setParameter("table", dsName);
	setParameter("withfilteredset", true);
	setParameter("filteredset", filteredSetName);
	setParameter("keepfilteroutput", true);
	setParameter("expression", expression);
	if (!quicklook())
		setParameter("writedss", true);

	ParameterInfo* info = gParameters;
	ParamDialog dialog(info, xmmSelectServer->gui());
	if(dialog.exec()) {
		xmmSelectServer->gui()->statusMessage("Filtering table ...");
		if (run()) {
			xmmSelectServer->gui()->statusMessage("... done", 1000);
			QMessageBox::information(xmmSelectServer->gui(),
									 "Filtering finished",
									 "The table has been successfully\n"
									 "filtered with the given expression.",
									 "Ok");
		} else {
			dataSetServer->remove(filteredSetName);
			xmmSelectServer->gui()->clearStatus();
		}
	}

	return true;
} // EvselectLauncher::filterTable()


bool EvselectLauncher::extractImageAndDisplay(const string &dsName,
											  const string &expression,
											  const TableColumn *xColumn,
											  const TableColumn *yColumn)
{
	if (!isValidExpression(expression))
		return false;

	setParameter("table", dsName);
	setParameter("expression", expression);
	if (!quicklook()) {
		setParameter("writedss", true);
		setParameter("updateexposure", true);
	}

	//	if the name of the axis columns ends in ...X/...Y we assume the
	//	user wants sqaurepixels by default
	string xCol(xColumn->name()),
		   yCol(yColumn->name());
	setParameter("squarepixels", toupper(xCol[xCol.length()-1]) == 'X' &&
								 toupper(yCol[yCol.length()-1]) == 'Y');

	if (xmmSelectServer->gui()->showIntensityImage()) {
		//
		//	accumulate image and display with imgdisplay
		//
		setParameter("withimageset", true);
		setParameter("imageset", "image.ds");
		setParameter("xcolumn", xCol);
		setParameter("ycolumn", yCol);

		ParameterInfo* info = gParameters;
		ParamDialog dialog(info, xmmSelectServer->gui());
		if(dialog.exec()) {
			xmmSelectServer->gui()->statusMessage("Accumulating image ...");
			string imgName;
			if (run()) {
				imgName = DsplotLauncher::toAbsolutePath(
												stringParameter("imageset"));
				DataSet *set = dataSetServer->open(imgName, Dal::Read,
												   DataSetServer::Low);
				Array *array = set->array(0);
				if (array->hasAttribute("CRPIX1") &&
					array->hasAttribute("CRVAL1") &&
					array->hasAttribute("CDELT1") &&
					array->hasAttribute("CTYPE1") &&
					array->attribute("CTYPE1")->asString() != "RA---TAN") {
					//	linear WCS transformation
					_binX = array->attribute("CDELT1")->asReal();
					_withMinX = true;
					_minX = (1.-array->attribute("CRPIX1")->asReal())*_binX +
							array->attribute("CRVAL1")->asReal();
				} else {
					_binX = realParameter("ximagebinsize");
					_withMinX = false;
					if (booleanParameter("withxranges")) {
						_withMinX = true;
						_minX = realParameter("ximagemin");
					}
				}
				if (array->hasAttribute("CRPIX2") &&
					array->hasAttribute("CRVAL2") &&
					array->hasAttribute("CDELT2") &&
					array->hasAttribute("CTYPE2") &&
					array->attribute("CTYPE2")->asString() != "DEC--TAN") {
					//	linear WCS transformation
					_binY = array->attribute("CDELT2")->asReal();
					_withMinY = true;
					_minY = (1.-array->attribute("CRPIX2")->asReal())*_binY +
								array->attribute("CRVAL2")->asReal();
				} else {
					_binY = realParameter("yimagebinsize");
					_withMinY = false;
					if (booleanParameter("withyranges")) {
						_withMinY = true;
						_minY = realParameter("yimagemin");
					}
				}
				dataSetServer->close(set);

				xmmSelectServer->gui()->statusMessage("Launching viewer "
													"and loading image ...");

				TheImageViewer::Instance().display(imgName);
				xmmSelectServer->gui()->statusMessage("... done", 1000);
			} else
				dataSetServer->remove(imgName);
		} else
			xmmSelectServer->gui()->clearStatus();
	} else {
		//
		//	filter table and display scatter plot with dsplot
		//
		xmmSelectServer->gui()->statusMessage("Filtering table ...");
		string filteredSetName = tmpnam(0);
		setParameter("withfilteredset", true);
		setParameter("filteredset", filteredSetName);
		setParameter("destruct", true);
		setParameter("keepfilteroutput", true);

		if (run()) {
			xmmSelectServer->gui()->statusMessage("Launching viewer "
												  "and loading image ...");
			DsplotLauncher dsPlot;
			dsPlot.displayDataSet(stringParameter("filteredset"),
								  xCol, yCol, true);
			dataSetServer->remove(filteredSetName);
			xmmSelectServer->gui()->statusMessage("... done", 1000);
		} else
			xmmSelectServer->gui()->clearStatus();
	}

	return true;
} // EvselectLauncher::extractImageAndDisplay()



bool EvselectLauncher::extractSpectrumAndDisplay(
						 const string &evTabName, const string &expression,
						 const TableColumn *column,
						 bool runBackscale)
{
	if (!isValidExpression(expression))
		return false;

	setParameter("table", evTabName);
	setParameter("expression", expression);
	setParameter("withspectrumset", true);
	setParameter("spectrumset", "spectrum.ds");
	setParameter("energycolumn", column->name());
	if (!quicklook()) {
		setParameter("writedss", true);
		setParameter("updateexposure", true);
	}

	ParameterInfo* info = gParameters;
	ParamDialog dialog(info, xmmSelectServer->gui());
	if(dialog.exec()) {
		xmmSelectServer->gui()->statusMessage("Accumulating spectrum ...");
		string specSet;
		bool withzcolumn = false;
		if (run()) {
			_oneDbin = intParameter("spectralbinsize");
			_withOneDmin = booleanParameter("withspecranges");
			_oneDmin = realParameter("specchannelmin");
			specSet = stringParameter("spectrumset");
			withzcolumn = booleanParameter("withzcolumn");
			if (runBackscale) {
				string tmpSpecSet(tmpnam(NULL));
				dataSetServer->copy(specSet, tmpSpecSet);	//	make a copy in
															//	case backscale
															//	fails
				xmmSelectServer->gui()->statusMessage("Running backscale on "
						"extracted spectrum ...");

				Info::Task *backscale = task("backscale");
				backscale->setParameter("withbadpixcorr", true);
				string dsName(evTabName);
				string::size_type pos = evTabName.find(':');
				backscale->setParameter("badpixlocation",
							pos != string::npos ? dsName.erase(pos) : dsName);
				backscale->setParameter("spectrumset", specSet);
				//
				//	invoke backscale in a mode that is does not make use
				//	of the ODF
				//
				backscale->setParameter("useodfatt", false);
				bool runbackscale = true;
				if (_withBackscaleDiag) {
					ParamDialog dialog(backscale->parameters(),
									   xmmSelectServer->gui());
					runbackscale = dialog.exec();
				}
				if (runbackscale)
					try {
						backscale->exec();
					} catch(SasException::Exception e) {
						QMessageBox::information(xmmSelectServer->gui(),
							"Backscale error", e.message.c_str(), "Ok");
						xmmSelectServer->gui()->clearStatus();
						QMessageBox::warning(xmmSelectServer->gui(),
							"Possible wrong BACKSCAL value",
							"As a consequence of the preceding error\nthe "
							"`BACKSCAL' value of the extracted spectrum\n"
							"is likely to be wrong.", "Ok");
						dataSetServer->remove(specSet);
						dataSetServer->copy(tmpSpecSet, specSet);
					}
				unlink(tmpSpecSet.c_str());
			}
			xmmSelectServer->gui()->statusMessage("Displaying spectrum "
												  "...");
			try {
				DsplotLauncher dsPlot;
				dsPlot.displayDataSet(
					dsPlot.toAbsolutePath(specSet) + ":SPECTRUM", "CHANNEL",
					!withzcolumn ? "COUNTS" : "RATE.STAT_ERR");
			} catch(SasException::Exception e) {
				QMessageBox::information(xmmSelectServer->gui(),
					"Display error", e.message.c_str(), "Ok");
				xmmSelectServer->gui()->clearStatus();	
			} catch(...) {
				QMessageBox::information(xmmSelectServer->gui(),
					"Display error", "unknown", "Ok");
				xmmSelectServer->gui()->clearStatus();	
			}
			xmmSelectServer->gui()->statusMessage("... done", 1000);
		} else {
			unlink(specSet.c_str());
			xmmSelectServer->gui()->clearStatus();
			return false;
		}
	}

	return true;
} // EvselectLauncher::extractSpectrumAndDisplay()


bool EvselectLauncher::extractRatesCurveAndDisplay(
						 const string &dsName, const string &expression,
						 const TableColumn *column)
{
	if (!isValidExpression(expression))
		return false;

	setParameter("table", dsName);
	setParameter("expression", expression);
	setParameter("withrateset", true);
	setParameter("timecolumn", column->name());
	setParameter("rateset", "rates.ds");
	setParameter("maketimecolumn", true);
	setParameter("makeratecolumn", true);
	if (!quicklook()) {
		setParameter("writedss", true);
		setParameter("updateexposure", true);
	}

	ParameterInfo* info = gParameters;
	ParamDialog dialog(info, xmmSelectServer->gui());
	if(dialog.exec()) {
		xmmSelectServer->gui()->statusMessage("Accumulating rates curve ...");
		string rateSet;
		if (run()) {
			rateSet = stringParameter("rateset");
			_oneDbin = 1.;
			_withOneDmin = booleanParameter("withtimeranges");
			_oneDmin = realParameter("timemin");
			xmmSelectServer->gui()
								->statusMessage("Displaying rates curve ...");
			bool hasTimeCol = booleanParameter("maketimecolumn");
			try {
				DsplotLauncher dsPlot;
				dsPlot.displayDataSet(
					dsPlot.toAbsolutePath(rateSet) + ":RATE",
							   hasTimeCol
									? stringParameter("timecolumn")
									: "",
							   booleanParameter("makeratecolumn")
									? "RATE.ERROR"
									: "COUNTS",
								false, hasTimeCol);
			} catch(SasException::Exception e) {
				QMessageBox::information(xmmSelectServer->gui(),
					"Display error", e.message.c_str(), "Ok");
				xmmSelectServer->gui()->clearStatus();	
			}
			xmmSelectServer->gui()->statusMessage("... done", 1000);
		} else {
			dataSetServer->remove(rateSet);
			xmmSelectServer->gui()->clearStatus();
		}
	}

	return true;
} // EvselectLauncher::extractRatesCurveAndDisplay()


bool EvselectLauncher::extractHistogramAndDisplay(
						 const string &dsName, const string &expression,
						 const TableColumn *column)
{

	if (!isValidExpression(expression))
		return false;

	setParameter("table", dsName);
	setParameter("expression", expression);
	setParameter("withhistogramset", true);
	setParameter("histogramset", "histo.ds");
	setParameter("histogramcolumn", column->name());

	ParameterInfo* info = gParameters;
	ParamDialog dialog(info, xmmSelectServer->gui());
	if(dialog.exec()) {
		xmmSelectServer->gui()->statusMessage("Accumulating histogram ...");
		string histSet;
		if (run()) {
			histSet = stringParameter("histogramset");
			_oneDbin = 1.;
			_withOneDmin = true;
			_oneDmin = 0.;
			xmmSelectServer->gui()->statusMessage("Displaying histogram ...");
			DsplotLauncher dsPlot;
			dsPlot.displayDataSet(
					dsPlot.toAbsolutePath(histSet) + ":HISTO",
							  column->name(), "COUNTS");
			xmmSelectServer->gui()->statusMessage("... done", 1000);
		} else {
			dataSetServer->remove(histSet);
			xmmSelectServer->gui()->clearStatus();
		}
	}

	return true;
} // EvselectLauncher::extractAndDisplayHistogram()


bool EvselectLauncher::isValidExpression(const string &expression)
{
	if (!Evaluator::isValidExpression(
		DataFilter::expandSelectionExpression(expression,
							*xmmSelectServer->tableInquiry()->openTable())))
	{
		stringstream s;

		s<<"The given selection expression\n"<<expression
		 <<"\nis syntacically incorrect.";
		QMessageBox::information(xmmSelectServer->gui(),
								 "Syntax error", s.str().c_str(), "Ok");
		xmmSelectServer->tableInquiry()->closeTable();

		return false;
	}

	xmmSelectServer->tableInquiry()->closeTable();
	return true;
} // EvselectLauncher::isValidExpression()


bool EvselectLauncher::run()
{
	using namespace SelectError;

	bool res = true;
	selectErrorHandler.exceptions(true);
	try {
		task(evselect);
	} catch(SelectException e) {
		res = false;
		QMessageBox::information(xmmSelectServer->gui(), "Selection error",
								 e.text().c_str(), "Ok");
		xmmSelectServer->gui()->clearStatus();
	} catch(SasException::Exception e) {
		res = false;
		QMessageBox::information(xmmSelectServer->gui(), "Evselect error",
								 e.message.c_str(), "Ok");
		xmmSelectServer->gui()->clearStatus();
	}
	selectErrorHandler.oldHandling();

	return res;
} // EvselectLauncher::run()


EvselectLauncher::~EvselectLauncher() {}

//
//	------------------------( especget )--------------------------------------
//
EspecgetLauncher::EspecgetLauncher() : TaskLauncher("especget")
{
	openParameters(name());
} // EspecgetLauncher::EspecgetLauncher()


bool EspecgetLauncher::extractProducts(const std::string &tableName,
									   const std::string &srcExpr,
									   const std::string &bgExpr)
{
	setParameter("filestem", "stem");
	setParameter("srcexp", srcExpr);
	setParameter("backexp", bgExpr);
	setParameter("table", tableName);

	ParameterInfo* info = gParameters;
	ParamDialog dialog(info, xmmSelectServer->gui());
	bool res = true;
	if(dialog.exec()) {
		xmmSelectServer->gui()->statusMessage("Accumulating spectral "
											  "products...");
		if (run())
			xmmSelectServer->gui()->statusMessage("... done", 1000);
		else {
			xmmSelectServer->gui()->clearStatus();
			res = false;
		}
	}

	return res;
} // EspecgetLauncher::extractProducts()


bool EspecgetLauncher::run()
{
	bool res = true;
	try {
		task(especget);
	} catch(SasException::Exception e) {
		res = false;
		QMessageBox::information(xmmSelectServer->gui(), "Especget error",
								 e.message.c_str(), "Ok");
		xmmSelectServer->gui()->clearStatus();
	}

	return res;
} // EspecgetLauncher::run()


EspecgetLauncher::~EspecgetLauncher() {}



//
//	------------------------( eregionanalyse )--------------------------------
//
SourceRegionOptimizer::SourceRegionOptimizer()
	: TaskLauncher("eregionanalyse")
{
	openParameters(name());
} // SourceRegionOptimizer::SourceRegionOptimizer()


SourceRegionOptimizer::OptParam
	SourceRegionOptimizer::optimizeSourceCircleRadius(const string &imgName,
													  const string &srcExpr,
													  const GeoRegion &srcReg,
													  const string &bgExpr)
const
{
	OptParam param = { srcReg[0][0]->refPoint().x(),
					   srcReg[0][0]->refPoint().y(),
					   1.2*srcReg[0][0]->radii()[0],
					   100, true };

if (!getenv("EREGIONSTUB")) {
	setParameter("imageset", imgName);
	setParameter("srcexp", srcExpr);
	setParameter("backexp", bgExpr);

	ParameterInfo* info = gParameters;
	ParamDialog dialog(info, xmmSelectServer->gui());
	bool res = true;
	if(dialog.exec()) {
		xmmSelectServer->gui()->statusMessage("Optimizing source extraction "
											  "region ...");
		if (run())
			xmmSelectServer->gui()->statusMessage("... done", 1000);
		else {
			xmmSelectServer->gui()->clearStatus();
			res = false;
		}
	}

	OptParam paramn = { realParameter("xcentroid"), realParameter("ycentroid"),
					   realParameter("optradius"), realParameter("srccnts"),
					   res };
	param = paramn;
}

	return param;
} // SourceRegionOptimizer::optimizeSourceCircleRadius()


bool SourceRegionOptimizer::run() const
{
	bool res = true;
	try {
		task(eregionanalyse);
	} catch(SasException::Exception e) {
		res = false;
		QMessageBox::information(xmmSelectServer->gui(),
								 "EregionAnalyse error", e.message.c_str(),
								 "Ok");
		xmmSelectServer->gui()->clearStatus();
	}

	return res;
} // SourceRegionOptimizer::run()


SourceRegionOptimizer::~SourceRegionOptimizer() {}


//
//	-----------------------( ImageViewer )--------------------------
//
ImageViewer::ImageViewer() : _isUp(false), _imageName(""), _ds9(0)
{
} // ImageViewer::ImageViewer()


Ds9 &ImageViewer::ds9()
{
	if (!_ds9)
		_ds9 = new Ds9;

	return *_ds9;
} // ImageViewer::ds9()


void ImageViewer::display(const string &name)
{
	_imageName = name;
	ds9().displayImage(name);
	_isUp = true;
} // ImageViewer::display()


void ImageViewer::down()
{
} // ImageViewer::down()


ImageViewer::~ImageViewer()
{
	delete _ds9;
} // ImageViewer::~ImageViewer()


DsplotLauncher::DsplotLauncher() : TaskLauncher("dsplot")
{
} // DsplotLauncher::DsplotLauncher()


double DsplotLauncher::offsetX() const
{
	return _offsetX;
} // DsplotLauncher::offsetX()


void DsplotLauncher::displayDataSet(const string &name, const string &xcol,
									const string &ycol,
									bool scatter, bool offsetX)
{
	const char *dsname = tmpnam(NULL);

	openParameters("dstoplot");
	setParameter("table", name);

	if (!xcol.empty()) {
		setParameter("withx", true);
		setParameter("x", xcol);
		setParameter("index", false);
	} else
		setParameter("index", true);

	vector<string> ycolvec(1);
	ycolvec[0] = ycol;
	setParameter("withy", true);
	setParameter("y", ycolvec);

	setParameter("withscatter", scatter);
	setParameter("withoffsetx", offsetX);

	setParameter("output", "file");
	setParameter("outputfile", dsname);

	task(dstoplot);
	_offsetX = offsetX ? realParameter("offsetx") : 0.;

	//	spawn grace
	//	for the moment always kick of a new grace subprocess until
	//	we find out how to delete all active sets and read in a new
	//	set
	if (_grace && _grace->isUp()) {
		delete _grace;
		_grace = new Grace(dsname);
//		_grace->killAll();
//		_grace->readSets(dsname);
	} else {
		delete _grace;
		_grace = new Grace(dsname);
	}
} // DsplotLauncher::displayDataSet()


//
//	workaround for bug in grace: need to read regions multiple times to get
//	the most actual one
//
void DsplotLauncher::pollRegion(vector<string> &v) const
{
	assert(_grace);
	for (unsigned n=0; n<5; ++n) {
		v = _grace->regionSelection();
		if (v.size() > 0)
			break;
	}
} // DsplotLauncher::pollRegion()


vector<string> DsplotLauncher::getRegion() const
{
	vector<string> v;
	if (_grace) {
		pollRegion(v);
		if (v.size() == 0) {
			sleep(1);
			pollRegion(v);
		}
	}

	return v;
} // DsplotLauncher::getRegion()


RegionSpecification *DsplotLauncher::regionSpec(const string &s) const
{
	return new DsplotRegionSpecification(s);
} // DsplotLauncher::regionSpec()


void DsplotLauncher::down() const
{
	delete _grace;
	_grace = 0;
} // DsplotLauncher::down()

DsplotLauncher::~DsplotLauncher()
{
} // DsplotLauncher::~DsplotLauncher()
