//----------------------------------------------------------------------------
//
//                     XMM Scence Analysis System (SAS)
//                       (c) 1997 European Space Agency
//
//---------------------------------------------------------------------------
//
//	TaskLauncher.h:	--- classes related to launching tasks
//
//	DESCRIPTION:
//		Classes related to launching tasks
//
//	CHANGE HISTORY:
//		UL, 1998-10-22: 1.0.0: - first version
//		UL, 1999-01-29: 1.1.0: - minor fix to make it build under egcs
//		UL, 2000-05-09: 1.2.0: - minor fix
//		UL, 2002-09-18: 1.3.0: - added various task wrapper classes
//								 (SourceRegionOptimizer etc.)
//
//	AUTHOR:
//		U. Lammers (UL), Thu Oct 22 14:15:14 MET DST 1998
//
//	$Header: /usr/local/cvs/sasdev/task_devel/packages/xmmselect/src/TaskLauncher.h,v 1.20 2002/09/18 15:24:16 ulammers Exp $
//
#ifndef TASK_LAUNCHER_H
#define TASK_LAUNCHER_H

#include <string>
#include <vector>

#include <Singleton.h>

#include <Grace.h>
#include <Ds9.h>
#undef DEBUG
#include "TableInquiry.h"
#include "ExceptionErrorHandler.h"
#include "ExprTranslator.h"

#include <qapplication.h>

class TaskLauncher : public QObject {
	Q_OBJECT

public:
	TaskLauncher(const std::string &name);

	const std::string &name() const { return _name; }

	~TaskLauncher();

private:
	std::string _name;
};


class EvselectLauncher : public TaskLauncher {
	Q_OBJECT

public:
	EvselectLauncher();

	double oneDbin() const { return _oneDbin; }
	double oneDmin() const { return _oneDmin; }
	bool withOneDmin() const { return _withOneDmin; }

	double binX() const { return _binX; }
	double binY() const { return _binY; }
	double minX() const { return _minX; }
	double minY() const { return _minY; }

	bool withMinX() const { return _withMinX; }
	bool withMinY() const { return _withMinY; }
	bool quicklook() const { return _quicklook; }
	bool withBackscaleDiag() const { return _withBackscaleDiag; }

	~EvselectLauncher();

public slots:
	bool filterTable(const std::string &dsName, const std::string &expression,
					 const std::string &filteredSet);
	bool extractImageAndDisplay(
						 const std::string &dsName,
						 const std::string &expression,
						 const TableColumn *xColumn,
						 const TableColumn *yColumn);
	bool extractSpectrumAndDisplay(
						 const std::string &dsName,
						 const std::string &expression,
						 const TableColumn *column,
						 bool runBackscale);
	bool extractRatesCurveAndDisplay(
						 const std::string &dsName,
						 const std::string &expression,
						 const TableColumn *column);
	bool extractHistogramAndDisplay(
						 const std::string &dsName,
						 const std::string &expression,
						 const TableColumn *column);

private:
	bool isValidExpression(const std::string &expression);
		//	true if expression is syntactically correct
	bool run();
		//	true if evselect ran successfully

	bool	_quicklook, _withBackscaleDiag;
	double	_binX, _binY,
			_minX, _minY;
	bool	_withMinX, _withMinY;

	double	_oneDmin, _oneDbin;
	bool	_withOneDmin;
}; // EvselectLauncher

class SourceRegionOptimizer : public TaskLauncher {
	Q_OBJECT

public:
	SourceRegionOptimizer();
	~SourceRegionOptimizer();

	struct OptParam {
		//	optimal source extraction parameters
		double xOpt, yOpt,	//	position
			   rOpt,		//	circle radius
			   srcCnts;		//	photon counts in source circle
		bool   ok;			//	results ok?
	};

	OptParam optimizeSourceCircleRadius(const string &imgName,
										const string &srcExpr,
										const ::GeoRegion &srcreg,
										const string &bgExpr) const;

private:
	bool run() const;
		//	true if Especget ran successfully
}; // SourceRegionOptimizer


class EspecgetLauncher : public TaskLauncher {
	Q_OBJECT

public:
	EspecgetLauncher();
	~EspecgetLauncher();

public slots:
	bool extractProducts(const std::string &tableName,
						 const std::string &srcExpr,
						 const std::string &bgExpr);

private:
	bool run();
		//	true if Especget ran successfully
}; // EspecgetLauncher


class ImageViewer {
public:
	ImageViewer();
	~ImageViewer();

	Ds9 &ds9();
	void display(const std::string &name);
	bool isImageUp() { return _isUp; }
	std::string imageName() { return _imageName; }
	void down();

private:
	bool				_isUp;
	std::string			_imageName;
	Ds9					*_ds9;
}; // ImageViewer

typedef Loki::SingletonHolder<ImageViewer>	TheImageViewer;


class DsplotLauncher : public TaskLauncher {
	Q_OBJECT

public:
	DsplotLauncher();
	~DsplotLauncher();

	RegionSpecification *regionSpec(const std::string &s) const;

	void down() const;

	//
	//	set/define offset value
	//
	double offsetX() const;

	void displayDataSet(const std::string &,
						const std::string &, const std::string &,
						bool scatter = false, bool offsetX = false);
	std::vector<std::string> getRegion() const;

	static std::string toAbsolutePath(const std::string &name);
		//
		//	if <name> is not a file name with an absolute path, put
		//	the absolute path to the current working directory in front
		//	and return; otherwise return with <fname> directly
		//

private:
	void pollRegion(std::vector<std::string> &) const;

	static Grace*	_grace;
	static double	_offsetX;
}; // DsplotLauncher

#endif

