//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	XmmSelectGui.cc:	--- implementation of XmmSelectGui classes
//
//	CHANGE HISTORY:
//		UL, 1998-10-22: 1.0.0: - first version
//		UL, 1999-03-01: 2.0.0: - many upgrades
//		UL, 1999-05-06: 2.1.0: - several modifications to support for
//								 invocation of tabcalc task
//		UL, 1999-09-20: 2.2.0: - brought in line with selectlib-4.12 +
//								 imgdisplay-1.10
//		UL, 1999-09-21: 2.3.0: - brought in line with selectlib-4.14
//		UL, 1999-10-14: 2.4.0: - set inital height of center pane depending
//								 on value of 'centerpaneheight' parameter
//								 (fix SPR 848: GUI too big)
//		UL, 1999-10-22: 2.5.0: - fixed SPR 914 : Incorrect expression created
//								 added "Region" menu which has two check
//								 buttons for choosing between
//								 union/intersection as the two possible
//								 ways to combine multiple regions
//		UL, 1999-12-17: 2.6.0: - changed to conversion of ds9/grace regions
//								 into selectlib expression (no longer makes
//								 use of ExprTranslator class)
//		UL, 2000-02-08: 2.7.0: - readXyRegion() more robust w.r.t to unknown
//								 regions (fixes SPR 1086: task crashes when
//								 trying to read in 1D region)
//		UL, 2000-02-11: 2.8.0: - general overhaul of geometry management
//							   - added support for Histogram product of
//								 evselect
//							   - support for on-line help added
//							   - full table name is now passed to evselect
//								 extractor
//		UL, 2000-05-09: 2.9.0: - take potential offset of x-axis into account
//								 when converting 1-D regions into selection
//								 expressions
//		UL, 2000-05-10: 2.10.0:- detect DSS information in table and give
//								 user the possibility to import corresponding
//								 selection expression
//		UL, 2001-04-26: 2.11.0:- fixed SPR 1994: Error in region intersection
//		UL, 2002-05-17: 2.12.0:- close previously opened ds9 session upon exit
//		UL, 2002-07-11. 2.13.0:- added support for 'spectral products'
//								 generation via especget
//		UL, 2002-09-18: 2.14.0:- added support for eregionanalyse
//								 (experimental)
//		UL, 2002-11-13: 2.15.0:- various changes related to better
//								 especget/eregionalayse support
//
//	AUTHOR:
//		U. Lammers (UL), Thu Oct 22 14:24:39 MET DST 1998
//
//	$Header: /usr/local/cvs/sasdev/task_devel/packages/xmmselect/src/XmmSelectGui.cc,v 1.43 2002/09/18 15:24:16 ulammers Exp $
//

#include <sstream>
#include <qglobal.h>
#if defined(DEBUG)
	#undef DEBUG	//	macro name clash with Tracer
#endif
#include <qapplication.h>
#include <qmessagebox.h>
#include <qgroupbox.h>
#include <qpushbutton.h>
#include <qkeycode.h>
#include <qlayout.h>
#include <qlabel.h>
#include <qfont.h>
#include <qtooltip.h>
#include <qstatusbar.h>

#include "common.h"
#include "XmmSelectGui.h"
#include "ColumnSelector.h"
#include "ExprConstructor.h"
#include "ParamDialog.h"
#include "TaskLauncher.h"
#include "Ds9.h"
#include "Grace.h"
#include "DataSubspace.h"
#include "stringutil.h"
#include "param.h"
#include "errstr.h"
#include "OnDemandScrollView.h"
#include "DataSetBrowser.h"
#include "Filename.h"

#include "tabcalc.h"
#include "phasecalc.h"
#include "dsrm.h"

using namespace std;

//	constructor
XmmSelectGui::XmmSelectGui(const char *name)
  : SasTaskWindow(name),
	_columnScrollView(0), _columnVbox(0), _columnSelector(0),
	_imageUp(false), _binX(1.), _binY(1.),
    _withMinX(false), _withMinY(false),
    _minX(1.), _minY(1.),
	_oneDmin(0.), _oneDbin(1.), _withOneDmin(false),
	_xyplotUp(Nothing),
	_runBackscale(false),
	_runRegionAnalyse(true)
{
	_central = new QWidget(this);
	setCentralWidget(_central);

	QVBoxLayout* vbox = new QVBoxLayout(_central, 0);
	setupMenuBar();
	vbox->addWidget(setupExpressionWidget(_central), 0);
	vbox->addSpacing(10);
	vbox->addWidget(setupColumnBox(_central, 4), 0);
	vbox->addSpacing(10);
	vbox->addWidget(setupRegionSelection(), 0);
	vbox->addSpacing(10);
	vbox->addWidget(setupBottomButtonRow(), 0);

	connect(xmmSelectServer, SIGNAL(tableChanged()),
			this, SLOT(updateColumnBox()));

	this->resize(this->sizeHint().width(), 430 + 50 *
				xmmSelectServer->parameters()["centerpaneheight"]->intValue());
	this->show();

	string s = string("Welcome to ") + Info::Task::current()->name() +
			   string(" ") + Info::Task::current()->version();
	statusBar()->message(s.c_str(), 4000);
} // XmmSelectGui::XmmSelectGui()


void XmmSelectGui::presentDssImportDialog()
{
	//
	//	check if the table has DSS info
	//
	string expr;
	const string &dssExpr = xmmSelectServer->tableInquiry()->dssSelections();
	if (!dssExpr.empty() &&
		QMessageBox::information(this, "XmmSelect",
					"The table contains data subspace information.\n"
					"Do you want to see the corresponding selection "
					"expression?", "&Yes", "No") == 0)
		expr = '(' + dssExpr + ')';

	if (!expr.empty())
		_exprConstructor->manageSelectionExpressionDialog(expr,
							"The table has been filtered with the following"
							"selection expression.\nPlease note: Filtering "
							"with this expression again will have no "
							"effect.", "Table selection history");
} // XmmSelectGui::presentDssImportDialog()



void XmmSelectGui::statusMessage(const QString &msg) const
{
	this->statusBar()->message(msg);
	qApp->processEvents();
} // XmmSelectGui::statusMessage()


void XmmSelectGui::statusMessage(const QString &msg, unsigned time) const
{
	this->statusBar()->message(msg, time);
	qApp->processEvents();
} // XmmSelectGui::statusMessage()


void XmmSelectGui::clearStatus() const
{
	this->statusBar()->clear();
} // XmmSelectGui::clearStatus()


QMenuBar *XmmSelectGui::setupMenuBar()
{
	//
	//	File menu
	//
	QPopupMenu *file = new QPopupMenu;
	CHECK_PTR(file);
	QToolTip::add(file, "Load new table/exit Xmmselect");
	file->insertItem("&New table", this, SLOT(loadNewTable()), CTRL+Key_N);
	file->insertItem("&Reload table", xmmSelectServer, SLOT(reloadTable()),
					 CTRL+Key_R);
	file->insertSeparator();
	file->insertItem("E&xit", this, SLOT(quit()), CTRL+Key_X);

	//
	//	Column menu
	//
	QPopupMenu *column = new QPopupMenu;
	CHECK_PTR(column);
	QToolTip::add(column, "Add/delete column to/from table");
	column->insertItem("A&dd new column", this, SLOT(addAnyColumn()),
					   CTRL+Key_D);
	column->insertItem("Add &PHASE column", this, SLOT(addPhaseColumn()),
					   CTRL+Key_P);
	column->insertSeparator();
	column->insertItem("De&lete column", this, SLOT(deleteColumn()),
					   CTRL+Key_L);

	//
	//	Region menu
	//
	_region = new QPopupMenu;
	CHECK_PTR(_region);
	QToolTip::add(_region, "Change way multiple 2-D regions are combined");
	_unifyRegionsId = _region->insertItem("&Union", this,
									  SLOT(unifyRegions()), CTRL+Key_U);
	_intersectRegionsId =
		_region->insertItem("&Intersection", this,
									  SLOT(intersectRegions()), CTRL+Key_I);
	unifyRegions();

	//
	//	Viewer menu
	//
	_viewer = new QPopupMenu;
	CHECK_PTR(_viewer);
	QToolTip::add(_viewer, "Change product viewing parameters");

	_binnedIntensityImageId = _viewer->insertItem("&Binned intensity image",
												  this,
												  SLOT(binnedIntensityImage()),
												  CTRL+Key_B);
	_scatterPlotImageId = _viewer->insertItem("&Scatter plot image",
											  this, SLOT(scatterPlotImage()),
											  CTRL+Key_S);
	binnedIntensityImage();

	//
	//	Product Post-Processing
	//
	_productProcess = new QPopupMenu;
	CHECK_PTR(_productProcess);
	QToolTip::add(_productProcess, "Product post-processing options");
	_backscaleId = _productProcess->insertItem("ba&ckscale on spectra",
											   this,
											   SLOT(backscale()),
											   CTRL+Key_C);
	_eregionId = _productProcess->insertItem("e&regionanalyse on source image",
											 this,
											 SLOT(eregionanalyse()),
										     CTRL+Key_R);

	EvselectLauncher evselect;
	_runBackscale = !evselect.quicklook();
	_productProcess->setItemChecked(_backscaleId, _runBackscale);
	_productProcess->setItemChecked(_eregionId, _runRegionAnalyse);

	menuBar()->insertItem("&File", file, -1, 0);
	menuBar()->insertItem("&Column", column, -1, 1);
	menuBar()->insertItem("&Region", _region, -1, 2);
	menuBar()->insertItem("&Viewer", _viewer, -1, 3);
	menuBar()->insertItem("&Products", _productProcess, -1, 4);

	return menuBar();
} // XmmSelectGui::setupMenuBar()


void XmmSelectGui::unifyRegions()
{
	_unifyRegions = true;
	_region->setItemChecked(_unifyRegionsId, true);
	_region->setItemChecked(_intersectRegionsId, false);
} // XmmSelectGui::unifyRegions()


void XmmSelectGui::intersectRegions()
{
	_unifyRegions = false;
	_region->setItemChecked(_unifyRegionsId, false);
	_region->setItemChecked(_intersectRegionsId, true);
} // XmmSelectGui::intersectRegions()


void XmmSelectGui::binnedIntensityImage()
{
	_showIntensityImage = true;
	_viewer->setItemChecked(_binnedIntensityImageId, true);
	_viewer->setItemChecked(_scatterPlotImageId, false);
} // XmmSelectGui::binnedIntensityImage()


void XmmSelectGui::scatterPlotImage()
{
	_showIntensityImage = false;
	_viewer->setItemChecked(_binnedIntensityImageId, false);
	_viewer->setItemChecked(_scatterPlotImageId, true);
} // XmmSelectGui::binnedIntensityImage()


void XmmSelectGui::backscale()
{
	_runBackscale = !_runBackscale;
	_productProcess->setItemChecked(_backscaleId, _runBackscale);
} // XmmSelectGui::backscale()


void XmmSelectGui::eregionanalyse()
{
	_runRegionAnalyse = !_runRegionAnalyse;
	_productProcess->setItemChecked(_eregionId, _runRegionAnalyse);
} // XmmSelectGui::eregionanalyse()


QWidget *XmmSelectGui::setupExpressionWidget(QWidget *parent,
											 const string &expr)
{
	_exprConstructor = new ExpressionConstructor(parent,
												 "expression constructor",
												 expr);

	return _exprConstructor;
} // XmmSelectGui::setupExpressionWidget


QWidget *XmmSelectGui::setupColumnBox(QWidget *parent, unsigned minColumns)
{
	if (!_columnSelector) {
		//	on first invocation we need to define a group box for the
		//	columnSelector widget - this will be reused in case the
		//	table has changed
	 	_columnGroupBox = new QGroupBox(parent);  //_central);
	 	_columnGroupBox->setTitle("Column selection");
	}

	FileName fname(xmmSelectServer->tableInquiry()->name());
	string boxtitle = "Column selection [" + fname.basename() + "]";
	_columnGroupBox->setTitle(boxtitle.c_str());

	delete _columnVbox;
	delete _columnSelector;

	_columnScrollView = new OnDemandScrollView(_columnGroupBox);

	_columnSelector = new ColumnSelector(_columnScrollView->viewport(),
									     "column selector",
										 _exprConstructor,
										 *xmmSelectServer->tableInquiry());

	// Maximum initial size
	_columnScrollView->setMinimumHeight(40 * minColumns);
	_columnScrollView->setMaximumPreferredHeight(
		50 * xmmSelectServer->parameters()["centerpaneheight"]->intValue());
	_columnScrollView->setContents(_columnSelector);
	_columnScrollView->show();

	_columnVbox = new QVBoxLayout(_columnGroupBox, 10);
	_columnVbox->addSpacing(10);
	_columnVbox->addWidget(_columnScrollView);

	return _columnGroupBox;
} // XmmSelectGui::setupColumnBox()


QWidget *XmmSelectGui::setupRegionSelection()
{
	QGroupBox *gb = new QGroupBox(_central);
	gb->setTitle("Region selection");

	QBoxLayout *hbox = new QHBoxLayout(gb, 20);

	QPushButton *_image = new QPushButton("2D region", gb),
				*_xyplot = new QPushButton("1D region", gb);

	QToolTip::add(_image, "Get region specification from displayed image");
	QToolTip::add(_xyplot, "Get region specification from displayed x-y plot");

	_image->setMinimumSize(_image->sizeHint());
	_image->setMaximumHeight(1.5 * _image->sizeHint().height());
	_xyplot->setMinimumSize(_xyplot->sizeHint());
	_xyplot->setMaximumHeight(1.5 * _xyplot->sizeHint().height());

	connect(_xyplot, SIGNAL(released()), this, SLOT(readXyplotRegions()));
	connect(_image, SIGNAL(released()), this, SLOT(readImageRegions()));

	hbox->addWidget(_xyplot, 1);
	hbox->addWidget(_image, 1);

	return gb;
} // XmmSelectGui::setupRegionSelection()


QWidget *XmmSelectGui::setupBottomButtonRow()
{
	QGroupBox *gb = new QGroupBox(_central);
	gb->setTitle("Product selection");

	QBoxLayout *vbox = new QVBoxLayout(gb, 20),
			   *hbox1 = new QHBoxLayout(vbox),
			   *hbox2 = new QHBoxLayout(vbox);

	QPushButton *_evlist = new QPushButton("Filtered Table", gb),
				*_image = new QPushButton("Image", gb),
				*_histo = new QPushButton("Histogram", gb),
				*_spect = new QPushButton("OGIP Spectrum", gb),
				*_lc = new QPushButton("OGIP Rate Curve", gb),
				*_specprod =
					new QPushButton("OGIP Spectral Products", gb);

	QToolTip::add(_evlist, "Generate filtered table");
	QToolTip::add(_image, "Accumulate an image");
	QToolTip::add(_image, "Accumulate an image");
	QToolTip::add(_histo, "Accumulate a one-dimensional histogram");
	QToolTip::add(_spect, "Accumulate an OGIP spectrum");
	QToolTip::add(_lc, "Accumulate an OGIP rates curve");
	QToolTip::add(_specprod, "Accumulate source + background OGIP spectrum "
				  "+ ARF + RMF");

	_evlist->setMinimumSize(_evlist->sizeHint());
	_evlist->setMaximumHeight(1.5 * _evlist->sizeHint().height());

	_image->setMinimumSize(_image->sizeHint());
	_image->setMaximumHeight(1.5 * _image->sizeHint().height());

	_histo->setMinimumSize(_histo->size());
	_histo->setMaximumHeight(1.5 * _histo->sizeHint().height());

	_spect->setMinimumSize(_spect->size());
	_spect->setMaximumHeight(1.5 * _spect->sizeHint().height());

	_lc->setMinimumSize(_lc->sizeHint());
	_lc->setMaximumHeight(1.5 * _lc->sizeHint().height());

	_specprod->setMinimumSize(_specprod->size());
	_specprod->setMaximumHeight(1.5 * _specprod->sizeHint().height());

	connect(_evlist, SIGNAL(released()), this, SLOT(filter()));
	connect(_image, SIGNAL(released()),	this, SLOT(image()));
	connect(_histo, SIGNAL(released()),	this, SLOT(histogram()));
	connect(_spect, SIGNAL(released()),	this, SLOT(spectrum()));
	connect(_lc, SIGNAL(released()), this, SLOT(ratesCurve()));
	connect(_specprod, SIGNAL(released()), this, SLOT(spectralProducts()));

	hbox1->addWidget(_evlist, 5);
	hbox1->addWidget(_image, 5);
	hbox1->addWidget(_histo, 5);

	hbox2->addWidget(_spect, 5);
	hbox2->addWidget(_lc, 5);
	hbox2->addWidget(_specprod, 5);

	return gb;
} // XmmSelectGui::setupBottomRow()


XmmSelectGui::~XmmSelectGui()
{
	DsplotLauncher dl;
	dl.down();

	ImageViewer idl;
	idl.down();
} // XmmSelectGui::~XmmSelectGui()


void XmmSelectGui::addAnyColumn()
{
	//
	//	run tabcalc
	//
	openParameters("tabcalc");
	vector<string> tabV;
	tabV.push_back(xmmSelectServer->tableInquiry()->name());
	setParameter("tables", tabV);

	ParameterInfo* info = gParameters;
	ParamDialog dialog(info, this);
	if(dialog.exec()) {
		tabcalc();
		emit newTable(tabV[0]);
	}
} // XmmSelectGui::addAnyColumn()


void XmmSelectGui::addPhaseColumn()
{
	//
	//	run phasecalc
	//
	openParameters("phasecalc");
	vector<string> tabV;
	tabV.push_back(xmmSelectServer->tableInquiry()->name());
	setParameter("tables", tabV);

	ParameterInfo* info = gParameters;
	ParamDialog dialog(info, this);
	if(dialog.exec()) {
		phasecalc();
		emit newTable(tabV[0]);
	}
} // XmmSelectGui::addPhaseColumn()


void XmmSelectGui::deleteColumn()
{
	const TableColumn *col = _columnSelector->oneDprojectionAxis();
	if (!col) {
		QMessageBox::warning(this, "Select column", "Please select first a "
							 "column\nfor deletion", "Ok");
		return;
	}

	//
	//	run dsrm
	//
	openParameters("dsrm");
	vector<string> tabV;
	tabV.push_back(xmmSelectServer->tableInquiry()->name() + ':' +col->name());
	setParameter("objects", tabV);

	dsrm();
	emit newTable(tabV[0]);
} // XmmSelectGui::deleteColumn()


void XmmSelectGui::loadNewTable()
{
	DataSetBrowser browser(this, "Load new table");
	browser.setMultiSelection(false);
	browser.setSelectable(DalListViewable::TableType);
	if (browser.exec() == DataSetBrowser::Accepted)
		emit newTable(browser.selection()[0]);
} // XmmSelectGui::loadNewTable()


void XmmSelectGui::updateColumnBox()
{
	setupColumnBox(_central);
	updateGeometry();
} // XmmSelectGui::updateBox()


void XmmSelectGui::readImageRegions()
{
	if (!_columnSelector->twoDprojectionAxisX() ||
		!_columnSelector->twoDprojectionAxisY()) {
		QMessageBox::warning(this,
			"Select image columns",
			"You need to select two columns\nbefore a region can be "
			"read", "Ok");
		return;
	}

	Ds9 &ds9 = TheImageViewer::Instance().ds9();
	GeoRegion reg;
	if (_showIntensityImage)
		//	region from ds9
		reg = ds9.regionSelection(ImageDisplayer::IMAGE, "red");
	else {
		//	region from grace
		DsplotLauncher il;
		reg = Grace::toRegion(il.getRegion());
	}

	string expr;
	try {
		expr = reg.toSelectExpression(
						_columnSelector->twoDprojectionAxisX()->name(),
						_columnSelector->twoDprojectionAxisY()->name(),
						_binX, _binY,
						_withMinX ? _minX :
							_columnSelector->twoDprojectionAxisXmin() +
							.5*_binX,
						_withMinY ? _minY :
							_columnSelector->twoDprojectionAxisYmin() +
							.5*_binY);
	} catch (UnexpressionableRegion &e) {
		QMessageBox::warning(this, "Invalid region specification",
							 e.text().c_str(), "Ok");
		return;
	}

	if (!expr.empty())
		_exprConstructor->appendExpression(expr,
								   _unifyRegions ?
									ExpressionConstructor::OR_IF_SAME_SENDER :
									ExpressionConstructor::AND,
								   "REGION");
	else
		QMessageBox::warning(this,
			 "No region defined",
			 "Could not get a region specification - maybe you have not "
			 "defined one\nor no image has been accumulated.",
			 "Ok");
} // XmmSelectGui::readImageRegions()


void XmmSelectGui::readXyplotRegions()
{
	if (!_columnSelector->oneDprojectionAxis()) {
		QMessageBox::warning(this,
			"Select column",
			"You need to select a column\nbefore a 1-D region can be "
			"read", "Ok");
		return;
	}

	DsplotLauncher	xyplot;
	GeoRegion reg;
	try {
		reg = Grace::toRegion(xyplot.getRegion());
	} catch (...) {
		QMessageBox::warning(this,
			"Invalid region",
			"The 1-D region you defined in Grace is invalid.\nPlease "
			"verify again your choice.", "Ok");
			return;
	}

	string expr;
	if (reg.size() == 1) {	//	there should only be one component
		Component   &comp = reg[0];
		for (unsigned n=0; n<comp.size(); ++n) {
			string e;
			try {
				double min;
				if (_withOneDmin)
					min = _oneDmin;
				else if (xyplot.offsetX() != 0.)
					min = xyplot.offsetX();
				else
					min = _columnSelector->oneDprojectionAxisMin();

				e = comp[n]->toSelectExpression(
							_columnSelector->oneDprojectionAxis()->name(),
							_oneDbin, min);
			} catch (UnexpressionableRegion &e) {
				QMessageBox::warning(this,
									"Invalid region specification",
									e.text().c_str(), "Ok");
				return;
			}
			if (!e.empty()) {
				if (n)
					expr += (_unifyRegions ? "||" : "&&");
				expr += e;
			}
		}
	}
	reg.free();

	if (!expr.empty())
		_exprConstructor->appendExpression(expr,
								   _unifyRegions ?
									ExpressionConstructor::OR_IF_SAME_SENDER :
									ExpressionConstructor::AND,
								   "REGION");

	else
		QMessageBox::warning(this,
		  "No region defined",
		  "Could not get a region specification - maybe you have not defined "
		  "any yet.", "Ok");
} // XmmSelectGui::readXyplotRegions()


void XmmSelectGui::filter()
{
	string expr(_exprConstructor->expression());
	EvselectLauncher evselect;
	evselect.filterTable(xmmSelectServer->tableInquiry()->name(),
						 !expr.empty() ? expr : "true", "filtered.ds");
}

void XmmSelectGui::image()
{
	if (_columnSelector->twoDprojectionAxisX() &&
		_columnSelector->twoDprojectionAxisY()) {
		EvselectLauncher evselect;
		string expression(_exprConstructor->expression());
		if (expression.empty())
			expression = "true";

		if (evselect.extractImageAndDisplay(
							xmmSelectServer->tableInquiry()->name(),
						    expression,
						    _columnSelector->twoDprojectionAxisX(),
						    _columnSelector->twoDprojectionAxisY())) {
			_imageUp = true;
			_binX = evselect.binX();
			_binY = evselect.binY();
			_withMinX = evselect.withMinX();
			_minX = evselect.minX();
			_withMinY = evselect.withMinY();
			_minY = evselect.minY();
		} else
			QMessageBox::warning(this,
				"Image accumulation failure",
				"Failed to accumulate and display an image", "Ok");
	} else
		QMessageBox::warning(this,
			"Select image columns",
			"You need to select two columns\nbefore an image can be "
			"accumulated", "Ok");
} // XmmSelectGui::image()


void XmmSelectGui::spectrum()
{
	extractOneDproductAndDisplay(Spectrum);
} // XmmSelectGui::spectrum()


void XmmSelectGui::spectralProducts()
{
	if (!TheImageViewer::Instance().isImageUp()) {
		QMessageBox::warning(this,
			"No image accumulated",
			"No spatial image appears to have been accumulated yet.\n"
			"This is a prerequisite for generating spectral products.", "Ok");
		return;
	}

	if (!_showIntensityImage) {
		QMessageBox::warning(this, "No intensity image",
			"Can only import source and background regions from "
			"intensity image.", "Ok");
		return;
	}

	Ds9 &ds9 = TheImageViewer::Instance().ds9();
	GeoRegion reg = ds9.regionSelection(ImageDisplayer::LINEAR, "red");

	string xCol = _columnSelector->twoDprojectionAxisX()->name(),
		   yCol = _columnSelector->twoDprojectionAxisY()->name();

	assert(reg.size() == 2);	//	src + background region
	string selReg[2];
	for (unsigned n=0; n<2; ++n)
		try {
			selReg[n] = reg[n].toSelectExpression(xCol, yCol);
			string s = n==0 ? "source" : "background",
				   s1 = "No " + s + " region",
				   s2 = s1 + " was selected - cannot continue without one.";
			if (selReg[n].empty()) {
				QMessageBox::warning(this, s1.c_str(), s2.c_str(), "Ok");
				return;
			}
		} catch (UnexpressionableRegion &e) {
			QMessageBox::warning(this, "Invalid region specification",
								 e.text().c_str(), "Ok");
			return;
		}

	//	trim b/g region
	selReg[1].erase(0, 1);	//	delete '!'
	if (selReg[1].size()>2 && selReg[1][0] == '(' && selReg[1][1] == '(') {
		//	remove one level of parenthesis
		selReg[1].erase(0, 1);
		selReg[1].erase(selReg[1].length()-1, 1);
	}

	if (_runRegionAnalyse) {
		SourceRegionOptimizer srcRegOpt;
		SourceRegionOptimizer::OptParam p =
			srcRegOpt.optimizeSourceCircleRadius(
										TheImageViewer::Instance().imageName(),
										selReg[0], reg, selReg[1]);
		if (!p.ok)
			return;

		Component c;
		c.push_back(new FCircle(FPoint(p.xOpt, p.yOpt), p.rOpt));
		c.front()->color(ShapeBasics::WHITE);
		::GeoRegion optReg;
		optReg.push_back(c);
		ds9.displayRegion(ImageDisplayer::Region(optReg,
												 ImageDisplayer::LINEAR));

		stringstream s;
		s<<"The source region optimization yielded the following "
		   "parameters for\nthe extraction circle:\n"
		 <<"X\t:"<<p.xOpt<<endl
		 <<"Y\t:"<<p.yOpt<<endl
		 <<"radius\t:"<<p.rOpt<<endl
		 <<"counts\t:"<<p.srcCnts<<endl
		 <<"The optimized circle is shown in the color white in the ds9"
		   " window.\nHow do you want to proceed?";
		int i = QMessageBox::information(this, "Source region optimization",
			s.str().c_str(), "Use &optimized region", "Use &initial region",
			"&Start over", 0, 2);
		ds9.deleteRegions();
		if (i<2) {
			if (!i) {
				c.front()->color(reg[0][0]->color());
				reg[0] = c;
				selReg[0] = optReg.toSelectExpression(xCol, yCol);
			}
			ds9.displayRegion(ImageDisplayer::Region(reg,
													 ImageDisplayer::LINEAR));
		} else
			return;
	}

	EspecgetLauncher especget;
	if (!especget.extractProducts(xmmSelectServer->tableInquiry()->name(),
								  selReg[0], selReg[1])) {
		QMessageBox::warning(this,
							 "Spectral product accumulation failure",
							 "Failed to accumulate spectral products", "Ok");
		return;
	}

	//	all ok - display source and b/g spectrum
	xmmSelectServer->gui()->statusMessage("Displaying spectra ...");

	DsplotLauncher plotter;
	string fileStem = stringParameter("filestem");
	plotter.displayDataSet(plotter.toAbsolutePath(fileStem) +
						   "_src.ds:SPECTRUM", "CHANNEL", "COUNTS");

	xmmSelectServer->gui()->statusMessage("... done", 1000);
} // XmmSelectGui::spectralProducts()


void XmmSelectGui::histogram()
{
	extractOneDproductAndDisplay(Histo);
} // XmmSelectGui::histogram()


void XmmSelectGui::ratesCurve()
{
	extractOneDproductAndDisplay(Rates);
} // XmmSelectGui::ratesCurve()



void XmmSelectGui::extractOneDproductAndDisplay(XyPlotType type)
{
	if (_columnSelector->oneDprojectionAxis()) {
		EvselectLauncher evselect;
		string expression(_exprConstructor->expression());

		if (expression.empty())
			expression = "true";

		bool done = false;
		switch (type) {
			case Spectrum:
					done = evselect.extractSpectrumAndDisplay(
									xmmSelectServer->tableInquiry()->name(),
									expression,
									_columnSelector->oneDprojectionAxis(),
									_runBackscale);
					break;
			case Rates:
					done = evselect.extractRatesCurveAndDisplay(
									xmmSelectServer->tableInquiry()->name(),
									expression,
									_columnSelector->oneDprojectionAxis());
					break;
			case Histo:
					done = evselect.extractHistogramAndDisplay(
									xmmSelectServer->tableInquiry()->name(),
									expression,
									_columnSelector->oneDprojectionAxis());
					break;
			default:break;
		}
		_withOneDmin = evselect.withOneDmin();
		_oneDmin = evselect.oneDmin();
		_oneDbin = evselect.oneDbin();
		if (done)
			_xyplotUp = type;
		else
			QMessageBox::warning(this,
					"Product accumulation failure",
					"Failed to accumulate and display a "
					"spectrum, histogram or rates curve", "Ok");
	} else
		QMessageBox::warning(this,
			"Select a column",
			"You need to select a column\nbefore a spectrum/histogram or\n"
			"rates curve can be accumulated", "Ok");
} // XmmSelectGui::extractOneDproductAndDisplay()
